﻿//------------------------------------------------------------
// Copyright (c) Microsoft Corporation.  All rights reserved.
//------------------------------------------------------------

namespace Microsoft.Azure.Cosmos.Encryption
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Linq;
    using System.Net;
    using System.Threading;
    using System.Threading.Tasks;

    internal sealed class EncryptionSettings
    {
        private readonly Dictionary<string, EncryptionSettingForProperty> encryptionSettingsDictByPropertyName;

        private EncryptionSettings(string containerRidValue, IReadOnlyList<string> partitionKeyPath)
        {
            this.ContainerRidValue = containerRidValue;
            this.PartitionKeyPaths = partitionKeyPath;
            this.encryptionSettingsDictByPropertyName = new Dictionary<string, EncryptionSettingForProperty>();
            this.PropertiesToEncrypt = this.encryptionSettingsDictByPropertyName.Keys;
        }

        public string ContainerRidValue { get; }

        public IReadOnlyList<string> PartitionKeyPaths { get; }

        public IEnumerable<string> PropertiesToEncrypt { get; }

        public static Task<EncryptionSettings> CreateAsync(EncryptionContainer encryptionContainer, CancellationToken cancellationToken)
        {
            return InitializeEncryptionSettingsAsync(encryptionContainer, cancellationToken);
        }

        public EncryptionSettingForProperty GetEncryptionSettingForProperty(string propertyName)
        {
            this.encryptionSettingsDictByPropertyName.TryGetValue(propertyName, out EncryptionSettingForProperty encryptionSettingsForProperty);

            return encryptionSettingsForProperty;
        }

        public void SetRequestHeaders(RequestOptions requestOptions)
        {
            requestOptions.AddRequestHeaders = (headers) =>
            {
                headers.Add(Constants.IsClientEncryptedHeader, bool.TrueString);
                headers.Add(Constants.IntendedCollectionHeader, this.ContainerRidValue);
            };
        }

        private static Data.Encryption.Cryptography.EncryptionType GetEncryptionTypeForProperty(ClientEncryptionIncludedPath clientEncryptionIncludedPath)
        {
            return clientEncryptionIncludedPath.EncryptionType switch
            {
                EncryptionType.Deterministic => Data.Encryption.Cryptography.EncryptionType.Deterministic,
                EncryptionType.Randomized => Data.Encryption.Cryptography.EncryptionType.Randomized,
                _ => throw new ArgumentException($"Invalid encryption type {clientEncryptionIncludedPath.EncryptionType}. Please refer to https://aka.ms/CosmosClientEncryption for more details. "),
            };
        }

        private static async Task<EncryptionSettings> InitializeEncryptionSettingsAsync(EncryptionContainer encryptionContainer, CancellationToken cancellationToken)
        {
            cancellationToken.ThrowIfCancellationRequested();

            ContainerResponse containerResponse = await encryptionContainer.ReadContainerAsync();

            Debug.Assert(containerResponse.StatusCode == HttpStatusCode.OK, "ReadContainerAsync request has failed as part of InitializeEncryptionSettingsAsync operation. ");
            Debug.Assert(containerResponse.Resource != null, "Null resource received in ContainerResponse as part of InitializeEncryptionSettingsAsync operation. ");

            // set the Database Rid.
            string databaseRidValue = containerResponse.Resource.SelfLink.Split('/').ElementAt(1);

            // set the Container Rid.
            string containerRidValue = containerResponse.Resource.SelfLink.Split('/').ElementAt(3);

            IReadOnlyList<string> partitionKeyPaths = null;

#if ENCRYPTIONPREVIEW
            if (containerResponse.Resource.PartitionKeyPaths.Any())
            {
                partitionKeyPaths = containerResponse.Resource.PartitionKeyPaths;
            }
            else
#endif
            {
                partitionKeyPaths = new List<string> { containerResponse.Resource.PartitionKeyPath };
            }

            // set the ClientEncryptionPolicy for the Settings.
            ClientEncryptionPolicy clientEncryptionPolicy = containerResponse.Resource.ClientEncryptionPolicy;

            EncryptionSettings encryptionSettings = new EncryptionSettings(containerRidValue, partitionKeyPaths);

            if (clientEncryptionPolicy != null)
            {
                if (clientEncryptionPolicy.PolicyFormatVersion > Constants.SupportedClientEncryptionPolicyFormatVersion)
                {
                    throw new InvalidOperationException("This version of Microsoft.Azure.Cosmos.Encryption cannot be used with this container." +
                        " Please upgrade to the latest version of the same. Please refer to https://aka.ms/CosmosClientEncryption for more details.");
                }

                // for each of the unique keys in the policy Add it in /Update the cache.
                foreach (string clientEncryptionKeyId in clientEncryptionPolicy.IncludedPaths.Select(x => x.ClientEncryptionKeyId).Distinct())
                {
                    await encryptionContainer.EncryptionCosmosClient.GetClientEncryptionKeyPropertiesAsync(
                         clientEncryptionKeyId: clientEncryptionKeyId,
                         encryptionContainer: encryptionContainer,
                         databaseRid: databaseRidValue,
                         ifNoneMatchEtag: null,
                         shouldForceRefresh: false,
                         cancellationToken: cancellationToken);
                }

                // update the property level setting.
                foreach (ClientEncryptionIncludedPath propertyToEncrypt in clientEncryptionPolicy.IncludedPaths)
                {
                    Data.Encryption.Cryptography.EncryptionType encryptionType = GetEncryptionTypeForProperty(propertyToEncrypt);

                    EncryptionSettingForProperty encryptionSettingsForProperty = new EncryptionSettingForProperty(
                        propertyToEncrypt.ClientEncryptionKeyId,
                        encryptionType,
                        encryptionContainer,
                        databaseRidValue);

                    string propertyName = propertyToEncrypt.Path.Substring(1);

                    encryptionSettings.SetEncryptionSettingForProperty(
                        propertyName,
                        encryptionSettingsForProperty);
                }
            }

            return encryptionSettings;
        }

        private void SetEncryptionSettingForProperty(
            string propertyName,
            EncryptionSettingForProperty encryptionSettingsForProperty)
        {
            this.encryptionSettingsDictByPropertyName[propertyName] = encryptionSettingsForProperty;
        }
    }
}